/*
 * Written by Oron Peled <oron@actcom.co.il>
 * Copyright (C) 2009, Xorcom
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <xtalk/proto_sync.h>
#include <xtalk/debug.h>
#include <psptalk.h>

static const char rcsid[] = "$Id$";

#define	DBG_MASK	0x02

#define	BIT(i)		(1 << (i))

#define	NUM_PS		4

/*
 * It's just an XTALK device with a USB transport
 */
struct psp_device {
	struct xtalk_base	*xtalk_base;
	struct xtalk_sync	*xtalk_sync;
	void			*transport_priv;
	uint8_t			control_byte;	/* current state */
	uint8_t			active_ps_mask;
	uint8_t			active_hub_mask;
	uint8_t			decivolts[NUM_PS];
};

CMD_DEF(PSP, LED_SEND,
	uint8_t	bitmask;
	);

CMD_DEF(PSP, HUB_STAT,
	);

CMD_DEF(PSP, HUB_STAT_REPLY,
	uint16_t	hubstate;
	);

CMD_DEF(PSP, PIC_ADRES,
	);

CMD_DEF(PSP, PIC_ADRES_REPLY,
	uint8_t	raw_voltage[NUM_PS];
	);

CMD_DEF(PSP, SET_MASK,
	uint16_t	raw_hub_mask;
	uint8_t		active_ps_mask;
	);

CMD_DEF(PSP, GET_MASK,
	);

CMD_DEF(PSP, GET_MASK_REPLY,
	uint16_t	raw_hub_mask;
	uint8_t		active_ps_mask;
	);

CMD_DEF(PSP, UPDATE_MASK,
	);

CMD_DEF(PSP, TWS_PORT_SET,
	uint8_t		portnum;
	);

CMD_DEF(PSP, TWS_PORT_GET,
	);

CMD_DEF(PSP, TWS_PORT_GET_REPLY,
	uint8_t		portnum;
	);

CMD_DEF(PSP, TWS_PWR_GET,
	);

CMD_DEF(PSP, TWS_PWR_GET_REPLY,
	uint8_t		port_mask;
	);

CMD_DEF(PSP, TWS_MODE_SET,
        uint8_t         wd_active;
        );

CMD_DEF(PSP, TWS_MODE_GET);
CMD_DEF(PSP, TWS_MODE_GET_REPLY,
        uint8_t         wd_active;
        );



union XTALK_PDATA(PSP) {
	MEMBER(PSP, LED_SEND);
	MEMBER(PSP, HUB_STAT);
	MEMBER(PSP, HUB_STAT_REPLY);
	MEMBER(PSP, PIC_ADRES);
	MEMBER(PSP, PIC_ADRES_REPLY);
	MEMBER(PSP, SET_MASK);
	MEMBER(PSP, GET_MASK);
	MEMBER(PSP, GET_MASK_REPLY);
	MEMBER(PSP, UPDATE_MASK);
	MEMBER(PSP, TWS_PORT_SET);
	MEMBER(PSP, TWS_PORT_GET);
	MEMBER(PSP, TWS_PORT_GET_REPLY);
	MEMBER(PSP, TWS_PWR_GET);
	MEMBER(PSP, TWS_PWR_GET_REPLY);
	MEMBER(PSP, TWS_MODE_SET);
	MEMBER(PSP, TWS_MODE_GET);
	MEMBER(PSP, TWS_MODE_GET_REPLY);
} PACKED members;

struct xtalk_protocol	psptalk_base = {
	.name	= "PSP",
	.proto_version = PSP_PROTO_VER,
	.commands = {
		CMD_SEND(PSP, LED_SEND),
		CMD_SEND(PSP, HUB_STAT),
		CMD_RECV(PSP, HUB_STAT_REPLY),
		CMD_SEND(PSP, PIC_ADRES),
		CMD_RECV(PSP, PIC_ADRES_REPLY),
		CMD_SEND(PSP, SET_MASK),
		CMD_SEND(PSP, GET_MASK),
		CMD_RECV(PSP, GET_MASK_REPLY),
		CMD_SEND(PSP, UPDATE_MASK),
		CMD_SEND(PSP, TWS_PORT_SET),
		CMD_SEND(PSP, TWS_PORT_GET),
		CMD_RECV(PSP, TWS_PORT_GET_REPLY),
		CMD_SEND(PSP, TWS_PWR_GET),
		CMD_RECV(PSP, TWS_PWR_GET_REPLY),
		CMD_SEND(PSP, TWS_MODE_SET),
		CMD_SEND(PSP, TWS_MODE_GET),
		CMD_RECV(PSP, TWS_MODE_GET_REPLY),
	},
	.ack_statuses = {
	}
};

/*
 * Convenience wrappers
 */
static struct xtalk_command *psp_new_command(
	struct psp_device *pspdev,
	uint8_t op,
	uint16_t extra_data)
{
	struct xtalk_command *cmd;

	if((cmd = new_command(pspdev->xtalk_base, op, extra_data)) == NULL) {
		ERR("new_command failed\n");
		return NULL;
	}
	return cmd;
}

static int psp_process_command(
	struct psp_device *pspdev,
	struct xtalk_command *cmd,
	struct xtalk_command **reply_ref)
{
	uint16_t sequence_number;
	int ret;

	ret = process_command(pspdev->xtalk_sync,
			cmd, reply_ref, &sequence_number);
	if(ret < 0) {
		ERR("process_command failed: sequence_number=%d ret=%d\n",
			sequence_number, ret);
		return ret;
	}
	return 0;
}

int psp_led(struct psp_device *pspdev, uint8_t bitmask)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	int			ret;

	DBG("\n");
	assert(pspdev != NULL);
	if((cmd = psp_new_command(pspdev, PSP_LED_SEND, 0)) == NULL)
		return -ENOMEM;
	CMD_FIELD(cmd, PSP, LED_SEND, bitmask) = bitmask & 0x7;
	ret = psp_process_command(pspdev, cmd, &reply);
	if(ret < 0)
		goto out;
out:
	free_command(reply);
	return ret;
}

uint16_t encode_hub_mask(uint8_t hub_mask)
{
	uint16_t	raw_hub_mask;
	int		i;

	raw_hub_mask = 0;
	for (i = 0; i < 8; i++) {
		if (hub_mask & (1 << i)) {
			raw_hub_mask |= (1 << (2 * i));
		}
	}
	/* workaround -- reverse endianes */
	raw_hub_mask = ((raw_hub_mask & 0xFF) << 8) | (0xFF & (raw_hub_mask >> 8));
	return raw_hub_mask;
}

uint8_t decode_hub_mask(uint16_t raw_hub_mask)
{
	uint8_t	hub_mask;
	int	i;

	/* workaround -- reverse endianes */
	raw_hub_mask = ((raw_hub_mask & 0xFF) << 8) | (0xFF & (raw_hub_mask >> 8));
	hub_mask = 0;
	for (i = 0; i < 8; i++) {
		int	conn = (raw_hub_mask >> (2 * i)) & 0x01;
		//printf("%d: %d\n", i, conn);
		hub_mask |= (conn << i);
	}
	return hub_mask;
}

/*
 * Using this command with wrong pspdev->active_hub_mask
 * may report random data, since firmware does not know
 * which USB hubs actually exist on the board.
 */
int psp_hub_stat(struct psp_device *pspdev)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	uint16_t		hubstate;
	uint8_t			hub_mask;
	uint8_t			hub_conn;
	int			ret;

	DBG("\n");
	assert(pspdev != NULL);
	hub_mask = pspdev->active_hub_mask;
	if((cmd = psp_new_command(pspdev, PSP_HUB_STAT, 0)) == NULL)
		return -ENOMEM;
	ret = psp_process_command(pspdev, cmd, &reply);
	if(ret < 0)
		goto out;
	hubstate = CMD_FIELD(reply, PSP, HUB_STAT_REPLY, hubstate);
	hub_conn = decode_hub_mask(hubstate);
	hub_conn &= hub_mask;
	ret = hub_conn;
	DBG("hub_mask=0x%X hubstate=0x%04X ret=%d\n", hub_mask, hubstate, ret);
out:
	free_command(reply);
	return ret;
}

int psp_voltage(struct psp_device *pspdev, uint8_t ps_num)
{
	DBG("\n");
	assert(pspdev != NULL);
	return pspdev->decivolts[ps_num];
}

/*
 * Returns value in Deci-Volts
 */
int psp_refresh_volts(struct psp_device *pspdev)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	uint8_t			*raw_voltage;
	double			voltage;
	int			ret;
	int			i;

	DBG("\n");
	assert(pspdev != NULL);
	if((cmd = psp_new_command(pspdev, PSP_PIC_ADRES, 0)) == NULL)
		return -ENOMEM;
	ret = psp_process_command(pspdev, cmd, &reply);
	if(ret < 0)
		goto out;
	raw_voltage = CMD_FIELD(reply, PSP, PIC_ADRES_REPLY, raw_voltage);
	for (i = 0; i < NUM_PS; i++) {
		voltage = raw_voltage[i] * UNIT_VOLT;
		voltage += VOLT_OFFS;
		voltage /= 10000;	/* 1/10 volts */
		pspdev->decivolts[i] = (int)(voltage * 10 + 0.5);
		DBG("i=%d: raw_voltage=%d decivolts=%d\n", i, raw_voltage[i], pspdev->decivolts[i]);
	}
	ret = 0;
out:
	free_command(reply);
	return ret;
}

int psp_set_mask(struct psp_device *pspdev, uint8_t ps_mask, uint8_t hub_mask)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	int			ret;

	assert(pspdev != NULL);
	ps_mask &= 0x0F;
	hub_mask &= 0xFF;
	if((cmd = psp_new_command(pspdev, PSP_SET_MASK, 0)) == NULL)
		return -ENOMEM;
	CMD_FIELD(cmd, PSP, SET_MASK, active_ps_mask) = ps_mask;
	CMD_FIELD(cmd, PSP, SET_MASK, raw_hub_mask) = encode_hub_mask(hub_mask);
	ret = psp_process_command(pspdev, cmd, &reply);
	if(ret < 0)
		goto out;
	pspdev->active_ps_mask = ps_mask;
	pspdev->active_hub_mask = hub_mask;
	DBG("ps_mask=0x%X hub_mask=0x%X\n", pspdev->active_ps_mask, pspdev->active_hub_mask);
out:
	free_command(reply);
	return ret;
}

int psp_get_mask(struct psp_device *pspdev)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	int			ret;
	uint16_t		raw_hub_mask;

	DBG("\n");
	assert(pspdev != NULL);
	if((cmd = psp_new_command(pspdev, PSP_GET_MASK, 0)) == NULL)
		return -ENOMEM;
	ret = psp_process_command(pspdev, cmd, &reply);
	if(ret < 0)
		goto out;
	pspdev->active_ps_mask = CMD_FIELD(reply, PSP, GET_MASK_REPLY, active_ps_mask);
	raw_hub_mask = CMD_FIELD(reply, PSP, GET_MASK_REPLY, raw_hub_mask);
	pspdev->active_hub_mask = decode_hub_mask(raw_hub_mask);
	DBG("ps_mask=0x%X hub_mask=0x%X\n",
		pspdev->active_ps_mask, pspdev->active_hub_mask);
out:
	free_command(reply);
	return ret;
}

int psp_update_mask(struct psp_device *pspdev)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	int			ret;

	assert(pspdev != NULL);
	if((cmd = psp_new_command(pspdev, PSP_UPDATE_MASK, 0)) == NULL)
		return -ENOMEM;
	ret = psp_process_command(pspdev, cmd, &reply);
	if(ret < 0)
		goto out;
	DBG("\n");
out:
	free_command(reply);
	return ret;
}

int psp_ps_mask(struct psp_device *pspdev)
{
	return pspdev->active_ps_mask;
}

int psp_hub_mask(struct psp_device *pspdev)
{
	return pspdev->active_hub_mask;
}

int psp_tws_port_set(struct psp_device *pspdev, int portnum)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	int			ret;

	DBG("\n");
	assert(pspdev != NULL);
	if(portnum < 0 || portnum > 1) {
		ERR("Invalid portnum=%d\n", portnum);
		return -EINVAL;
	}
	if((cmd = psp_new_command(pspdev, PSP_TWS_PORT_SET, 0)) == NULL)
		return -ENOMEM;
	CMD_FIELD(cmd, PSP, TWS_PORT_SET, portnum) = portnum;
	ret = psp_process_command(pspdev, cmd, &reply);
	if(ret < 0)
		goto out;
	DBG("portnum=%d\n", ret);
out:
	free_command(reply);
	return ret;
}

int psp_tws_port_get(struct psp_device *pspdev)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	int			ret;

	DBG("\n");
	assert(pspdev != NULL);
	if((cmd = psp_new_command(pspdev, PSP_TWS_PORT_GET, 0)) == NULL)
		return -ENOMEM;
	ret = psp_process_command(pspdev, cmd, &reply);
	if(ret < 0)
		goto out;
	ret = CMD_FIELD(reply, PSP, TWS_PORT_GET_REPLY, portnum);
	DBG("portnum=%d\n", ret);
out:
	free_command(reply);
	return ret;
}

int psp_tws_pwr_get(struct psp_device *pspdev)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	int			ret;

	DBG("\n");
	assert(pspdev != NULL);
	if((cmd = psp_new_command(pspdev, PSP_TWS_PWR_GET, 0)) == NULL)
		return -ENOMEM;
	ret = psp_process_command(pspdev, cmd, &reply);
	if(ret < 0)
		goto out;
	ret = CMD_FIELD(reply, PSP, TWS_PWR_GET_REPLY, port_mask);
	DBG("port_mask=0x%X\n", ret);
out:
	free_command(reply);
	return ret;
}

int psp_tws_mode_set(struct psp_device *pspdev, int enable)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	int			ret;

	assert(pspdev != NULL);
	if((cmd = psp_new_command(pspdev, PSP_TWS_MODE_SET, 0)) == NULL)
		return -ENOMEM;
	CMD_FIELD(cmd, PSP, TWS_MODE_SET, wd_active) = (enable) ? 1 : 0;
	ret = psp_process_command(pspdev, cmd, &reply);
	if(ret < 0)
		goto out;
	DBG("%s\n", (enable) ? "set" : "unset");
out:
	free_command(reply);
	return ret;
}

int psp_tws_mode_get(struct psp_device *pspdev)
{
	struct xtalk_command	*cmd;
	struct xtalk_command	*reply;
	int			ret;

	DBG("\n");
	assert(pspdev != NULL);
	if((cmd = psp_new_command(pspdev, PSP_TWS_MODE_GET, 0)) == NULL)
		return -ENOMEM;
	ret = psp_process_command(pspdev, cmd, &reply);
	if(ret < 0)
		goto out;
	ret = CMD_FIELD(reply, PSP, TWS_MODE_GET_REPLY, wd_active);
	DBG("watchdog=%d\n", ret);
out:
	free_command(reply);
	return ret;
}

/*
 * Wrappers
 */

struct psp_device *psp_new(struct xusb_iface *iface)
{
	struct psp_device	*pspdev;
	struct xtalk_device	*xtalk_dev;
	int			ret;

	DBG("\n");
	pspdev = calloc(sizeof(*pspdev), 1);
	if(!pspdev) {
		ERR("Allocating PCP device memory failed\n");
		goto err;
	}
	pspdev->xtalk_base = xtalk_base_new_on_xusb(iface);
	pspdev->xtalk_sync = xtalk_sync_new(pspdev->xtalk_base);

	ret = xtalk_sync_set_protocol(pspdev->xtalk_sync, &psptalk_base);
	if(ret < 0) {
		ERR("PSP Protocol registration failed: %d\n", ret);
		goto err;
	}
	ret = xtalk_proto_query(pspdev->xtalk_sync);
	if(ret < 0) {
		ERR("Protocol handshake failed: %d\n", ret);
		goto err;
	}
	ret = psp_get_mask(pspdev);
	if(ret < 0) {
		ERR("psp_get_mask failed: %d\n", ret);
		goto err;
	}
	return pspdev;

err:
	if (pspdev)
		psp_delete(pspdev);
	return NULL;
}

void psp_delete(struct psp_device *pspdev)
{
	DBG("\n");
	if(pspdev) {
		xtalk_sync_delete(pspdev->xtalk_sync);
		pspdev->xtalk_base = NULL;
		free(pspdev);
	}
}
